open util/integer

one sig Bottom {
	v: one Int
} {
	v=-99
}

sig Thread {}
sig MemoryLocation {}

/* =============================== */
/* === Simplified RMA encoding === */
/* =============================== */

sig Item {}

abstract sig Action extends Item {
	po, hb, co : set Action, /* program order, happens before, consistency order */
	so : set Action,
	o, d : one Thread	/* origin and destination thread of the action */
} 

/* ===================== */
/* === Memory action === */
/* ===================== */

abstract sig MA extends Action {
	rl, wl : lone MemoryLocation, /* read location, write location */
	rV, wV: lone (Int + Bottom) /* value, read or written */
}

sig Reader in MA {
	W: one Writer /* operation who wrote the value written by this read */
} 

// one sig InitWrite {} /* initial memory state? */

sig Writer in MA {}

sig LocalWriters in MA{}

sig LocalReaders in MA{}

sig NicReaders in MA{}

sig NicWriters in MA {}


abstract sig RemoteAction extends MA {
	virtualOp: one VA
}

abstract sig RPG extends RemoteAction {} /* set of Remote Put and Remote Get */

abstract sig RA extends RemoteAction {} /* set of Remote Get Accumulate and Remote CAS */

abstract sig VA extends MA {} {/* set of virtual actions */
	o = d
}

/* ===== local actions ===== */
sig Write extends MA {} {
	o = d
	no so
	one wl
	no rl
	no rV
}

sig Read extends MA {}  {
	o = d
	no so
	one rl
	no wl
	no wV
}

/* ===== Remote Put ===== */
sig RemotePut extends RPG {} {
	virtualOp in VirtualPut
	one rl
	no wl
	no wV
}

sig VirtualPut extends VA {} {
	one wl
	no rl
	no rV
}

/* ===== Remote Get ===== */

sig RemoteGet extends RPG {} {
	virtualOp in VirtualGet
	one wl
  no rl
	no rV
}

sig VirtualGet extends VA {} {
	one rl
	no wl
	no wV
}

/* ===== RGA ===== */

sig RemoteGetAccumulate extends RA {} {
	virtualOp in VirtualRGA
	one wl
	no rl
	no rV
}

sig VirtualRGA extends VA {} {
	one wl
	one rl
	rl = wl
	one rV
	one wV
} 

/* ===== CAS ===== */

sig RemoteCAS extends RA {
	// EVENTUAL second read
	rl2 : lone MemoryLocation,
	rV2 : lone (Int + Bottom),
	W2:	lone Writer
} {
	virtualOp in VirtualCAS
	no wl
	no wV
}

sig VirtualCAS extends VA {
} {
	one rl
	one rV
	/* EVENTUAL write*/
	lone wl
	lone wV
} 


/* ============================== */
/* === Synchronization action === */
/* ============================== */

abstract sig SA extends Action {}

sig Flush extends SA {} {}

/* ============= */
/* === RULES === */
/* ============= */

fact { 
	po = ^po
}

fact {
	co = ^co
}

fact {
	 all a: Action | 
		not a in po[a]  /* no cycles in program order */
}


fact {
	all a: Action | 
		not a in co[a] /* consistency order is not reflexive */
}

fact { /* program order only between threads? */
	all a, b: Action | 
		b in po[a] implies o[a] = o[b] 
}

/* NOT ADDED */
fact { /* every two actions in the same thread are ordered by po */
	all disj a, b: Action | 
		o[a] = o[b] implies (a in po[b] or b in po[a])  
}

fact { /* so relation only appears inter threads */
	all a, b: Action | 
		b in so[a] implies o[a] != o[b]
}

fact {
 	hb = ^(po + so)
}


//TODO: add in the document
fact {
	all r : Reader | 
		rl[r] = wl[W[r]]
}

fact local_readers {
	LocalReaders = Read
}

fact nic_readers {
	NicReaders = 
		VirtualGet + RemotePut +
		VirtualRGA + RemoteCAS +
		VirtualCAS
}

/* all reader actions */
fact readers {
	Reader = 
		LocalReaders + NicReaders
}

fact local_writers {
	LocalWriters = Write
}

fact nic_writers {
	NicWriters = 
		VirtualPut + RemoteGet +
		RemoteGetAccumulate +
		VirtualCAS + VirtualRGA
}

/* all writer actions */
fact writers {
	Writer = 
		LocalWriters + NicWriters
}

fact {
	all r: RemoteAction | 
		/* virtual action happens in the destination thread */
				d[r.virtualOp] = d[r] and 
				o[r.virtualOp] = d[r]
}

fact {
	all r: RemotePut | 
				wV[r.virtualOp] = rV[r] 
}

fact {
	all r: RemoteGet | 
				rV[r.virtualOp] = wV[r] 
}

fact {
	all r: RemoteGetAccumulate | 
				rV[r.virtualOp] = wV[r]
}

//TODO: check if this is expected behaviour
fact {
	all v: VirtualRGA |
		(rV[v] = Bottom implies wV[v] = Bottom) and
		((not rV[v] = Bottom) implies wV[v] = rV[v].plus[1])
}

fact {
	all disj r: RemoteCAS, w: Writer |
		(rV[r] = rV[r.virtualOp] implies 
			/* second read in RemoteCAS and write in VirtualCAS */
			(wV[r.virtualOp] = rV2[r] and

			 one rV2[r] and one rl2[r] and one W2[r] and
			 one wl[r.virtualOp] and one wV[r.virtualOp] and 
			 wl[W2[r]] = rl2[r] and
			 cohb[W2[r], r] and rV2[r] = wV[W2[r]] and o[r] = o[W2[r]] and
			 ((cohb[W2[r], w] and cohb[w, r]) implies rl2[r] != wl[w]) and
			 not cohb[r, W2[r]])) 

		and
		
		(not (rV[r] = rV[r.virtualOp]) implies 
			/* NO second read in RemoteCAS and NO write in VirtualCAS */
			(no wV[r.virtualOp] and no wl[r.virtualOp] and
			 no rV2[r] and no rl2[r] and no W2[r]))
}

//sig ExternalSyncSend extends SA {
//	op: one ExternalSyncRecv
//}

//sig ExternalSyncRecv extends SA {
//	op: one ExternalSyncSend
//}


/* ======================== */
/* === Valid Executions === */
/* ======================== */

/* hb is acyclic (excluding fences, unlocks and flushes) */
fact e {
	all a: Action | 
				not a in hb[a]
}

/* ================================= */
/* === Epochs and sync semantics === */
/* ================================= */

/* hb (d, {r, w}) ⇒ co (d, {r, w}) (17) */
/* hb ({r, w}, d) ⇒ co ({r, w}, d)  (18) */
fact fg {
	all a: Write + Read, b: VA | 
					(a in hb[b] implies a in co[b]) and
					(b in hb[a] implies b in co[a])
}


/* po ({ra, rp, r, w}, fl) ⇒ co ({ra, rp, r, w}, fl)  (20) */
fact i {
	all a: RA + RemotePut + Read + Write, b: Flush |
			a in po[b] implies a in co[b]
}


fact j1 {
	all a: Read + Write, b: Flush | 
			b in po[a] implies b in co[a]
}

fact j2 {
	all a: RA + RemotePut, b: Flush |
			b in po[a] and d[a] = d[b] 
				implies b in co[a]
}

/* hb (d, {ca, d}) ⇒ co (d, {ca, d})  (23) */
/*fact l {
	all a: VA, b: RPG + RA + VA|
			b in a.hb implies b in a.co
}*/

/* po ({r, w}, ca) ⇒ co ({r, w}, ca)  (24) */
fact m {
	all a: Read + Write, b: RemoteAction |
			b in po[a] 
				implies b in co[a]
}

fact new1 {
	all a: RemoteAction | 
		a.virtualOp in co[a]
}

fact {
	all a: RemoteAction | 
		a.virtualOp in so[a]
}

fact new2 {
	all a: RemoteAction, b:Flush | 
		b in po[a] 
			implies (b in co[a.virtualOp] and b in so[a.virtualOp])
}



/* ============================================== */
/* === Rules for consistent memory operations === */
/* ============================================== */

pred conflicting[a, b: Action] {
	//0 = 1 and
	d[a] = d[b] and
	(
		// one of the two operations is a remote put 
		(a in VirtualPut and (wl[a] = wl[b] or wl[a] = rl[b])) or
		// exactly one of the operations is an accumulate action (RemoteCAS or RemoteGetAccumulate)
		(a in VirtualCAS and (wl[a] = wl[b] or wl[a] = rl[b])) or
		(a in VirtualRGA and (wl[a] = wl[b] or wl[a] = rl[b])) or
		(a in RemoteGetAccumulate and (wl[a] = wl[b] or wl[a] = rl[b])) or
		// one operation is a RemoteGet and the second one is a local write 
		(a in VirtualGet and b in Write and rl[a] = wl[b])
	)
}

/*
pred conflicting[a, b: Action] {
	d[a] = d[b] and
	(
	(a in RemotePut and 
		((b in Read and 
			wl[a.virtualOp] = rl[b]) or 

		 (b in RemoteGet and
			wl[a.virtualOp] = rl[b.virtualOp])
		 ) 
	 ) or

	(a in RemotePut and
		((b in Write  and
			wl[a.virtualOp] = wl[b]) or

		(b in RemotePut and
			wl[a.virtualOp] = wl[b.virtualOp])
		 ) 
	 ) or


	(a in RA and 
		b in Write + RemotePut and
		((rl2[a] = wl[b]) or (rl[a] = wl[b]))) or

	(a in RemoteGetAccumulate and 
		b in Read + RemoteGet and
		wl[a] = rl[b]) or

	(a in RemoteCAS and 
		b in Read + RemoteGet and
		wl[a.virtualOp] = rl[b]) or

	(a in RemoteGet and 
		b in Write and
		rl[a.virtualOp] = wl[b])
	)
}*/

pred data_race[a, b: Action] {
	(conflicting[a, b] or conflicting[b, a]) and
	((not a in hb[b] and not b in hb[a]) or
	(not a in co[b] and not b in co[a]))
}

/* no data races */
fact no_data_races {
	no disj a, b: Action | 
		data_race[a, b]
}

pred cohb[a, b: Action] {
	b in hb[a] and b in co[a]
}

/* cohb( W(r), r ) and V( r ) = V( W(r) )  (28) */
fact f28 {
	all r: Reader | 
		cohb[W[r], r] and rV[r] = wV[W[r]] and o[r] = o[W[r]] // TODO: why not d[r] = d[W[r]]??
}

/* cohb( W(r), w ) and cohb( w, r ) ⇒ r.rl != w.wl and V( r ) = V( W(r) )  (29) */
fact f29 {
	all r: Reader, w: Writer |
			cohb[W[r], w] and cohb[w, r] 
				implies rl[r] != wl[w]
}

/* not  cohb( r, W( r ) )  (30) */
fact f30 {
	no r: Reader | 
		cohb[r, W[r]]
}

/* ============================= */
/* === Memory ordering rules === */
/* ============================= */

fact f31 {
	all a, b: RA | 
		(wl[a] = wl[b] and b in po[a]) 
			implies b in co[a]
}
