import java.io.*;

import java.util.List;
import java.util.ArrayList;

import org.antlr.v4.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTree;

public class ConvertToDmapp {
    public static String DEFAULT_OUTPUT_DIR = "output";

    String inputFileName;
    ParseTree tree;
    VarsValue varsValue;
    InstanceValue instanceValue;
    TranslateValue translateValue;

    public ConvertToDmapp(String inputFileName, ParseTree tree) {
        this.inputFileName = inputFileName;
        this.tree = tree;

        translateValue = new TranslateValue();
        TranslateVisitor translateVisitor = new TranslateVisitor(translateValue);
        translateVisitor.visit(tree);

        instanceValue = new InstanceValue();
        DmappVisitor dmappVisitor = new DmappVisitor(instanceValue);
        dmappVisitor.visit(tree);

        varsValue = new VarsValue();
        VarsVisitor varsVisitor = new VarsVisitor(varsValue, instanceValue);
        varsVisitor.visit(tree);
    }

    public void incorporateOriginalIR(PrintStream os) throws java.io.IOException {
        BufferedReader br = new BufferedReader(new FileReader(inputFileName));
        String s = null;
        while ((s = br.readLine()) != null) {
            os.println(s);
        }
        br.close();
    }

    public void incorporateVarDecls(PrintStream os) throws java.io.IOException {
        for (Proc p : instanceValue.processes) {
            for (Var v : varsValue.procToLocal.get(p)) {
                // (local vars don't have initial values)
                String rhs = varsValue.localToRHS.get(v);
                if (rhs == null) rhs = "-1L";
                os.printf("    long                %s = %s;\n", v, rhs);
            }
        }

        /* counters for the observed outputs of the program */
        for (int i = 0; i < instanceValue.outputs.size(); i++) {
            os.printf("    long                cnt_%d = 0;\n", i);
        }


        for (Proc p : instanceValue.processes) {
            for (Var v : varsValue.procToShared.get(p)) {
                os.printf("    long                *%s = NULL;\n", v);
            }
        }
    }

    public void incorporateVars(PrintStream os) throws java.io.IOException {
        for (Proc p : instanceValue.processes) {
            for (Var v : varsValue.procToShared.get(p)) {
                os.printf("\t%s = (long *)dmapp_sheap_malloc(sizeof(long));\n", v.name);
                os.printf("\tif (%s == NULL) {\n", v.name);
                os.printf("\t\tfprintf(stderr, \" sheap_malloc'd failed ? 0x%%lx\\n\", (long)%s);\n", v.name);
                os.printf("\t\texit(1);\n");
                os.printf("\t}\n");
		        if (v.hasInitialValue) {
                    os.printf("\t");
                    if (instanceValue.processes.size() > 1) {
                        os.printf("if ((pe %% %d) == %d) ",
                            instanceValue.processes.size(), p.procNumber);
                    }
		            os.printf(" *%s = %d;\n",
                            v.name, v.iv);
		        }
            }
        }
    }

    public void incorporateTestBody(PrintStream os) throws java.io.IOException {
        for (Proc p : instanceValue.processes) {
            if (instanceValue.processes.size() > 1) {
                os.printf("\tif ((pe %% %d) == %d) {\n",
                        instanceValue.processes.size(), p.procNumber);
                if (p.procNumber == 0) {
                    /* add long sleep for first thread */
                    os.println("\t\tif (rand() % 2 == 1) { usleep(5 + (rand()%10)); }");
                }
            }

            for (String s : instanceValue.processContents.get(p.procNumber)) {
                if (instanceValue.processes.size() > 1) {
                    os.printf("\t");
                }
                /* for each statement except error checking ones, add a sleep before it */
                if (!s.contains("DMAPP_RC_SUCCESS")) {
                    os.println("\tif (rand() % 2 == 1) { usleep(1); }");
                }
                if (instanceValue.processes.size() > 1) {
                    os.printf("\t");
                }
                os.println("\t" + s);
            }

            if (instanceValue.processes.size() > 1) {
                os.printf("\t}\n");
            }
        }
    }

    StringBuilder simpleExpressionBuffer;
    List<String> vars;
    private Object visitSimpleExpression(TLParser.SimpleExpressionContext ctx) {
        if (ctx.And() != null) {
            /* simpleExpression And simpleExpression*/
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" && ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Or() != null) {
            /* simpleExpression Or simpleExpression */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" || ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Excl() != null) {
            /* Excl simpleExpression */
            simpleExpressionBuffer.append("!(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        } else if (ctx.OParen() != null) {
            /* OParen simpleExpression CParen */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        }else if (ctx.Equals() != null) {
            /* Identifier Equals rhs */
            vars.add(ctx.Identifier().getText());
            simpleExpressionBuffer.append(ctx.Identifier().getText() + " == " );
            visitRhs(ctx.rhs());

        } else if (ctx.NEquals() != null) {
            /* Identifier NEquals rhs */
            vars.add(ctx.Identifier().getText());
            simpleExpressionBuffer.append("(" + ctx.Identifier().getText() + " != " );
            visitRhs(ctx.rhs());
            simpleExpressionBuffer.append(")");

        } else {
            System.err.println("Error");
            System.exit(1);
        }
        return null;
    }

    public Object visitRhs(TLParser.RhsContext ctx) {
        if (ctx.Number() != null) {
            simpleExpressionBuffer.append(ctx.getText()+"L");
        } else {
            simpleExpressionBuffer.append(ctx.getText());
        }
        return null;
    }

    private String checkThread(int nrProcs, int procId) {
        if (nrProcs == 1) {
            return "(1)";
        } else {
            return "((pe % " + nrProcs + ") == " + (procId) + ")";
        }
    }

    public void incorporateTestWitnesses(PrintStream os) throws java.io.IOException {
        boolean firstVar = true;
        StringBuilder varFormats = new StringBuilder(),
            varNames = new StringBuilder();

        for (Proc p : instanceValue.processes) {
            if (!varsValue.procToLocal.get(p).isEmpty()) {
                os.printf("\tif %s {\n",
                        checkThread(instanceValue.processes.size(), p.procNumber));
                for (Var v : varsValue.procToLocal.get(p)) {
                    // (local vars don't have initial values)
                    if (p.procNumber == 0) {
                        os.printf("\t\t*%s_receiver = %s;\n", v, v);
                    } else {
                        os.printf("\t\t*%s_sender = %s;\n", v, v);
                        os.printf("\t\tstatus = dmapp_put_nbi(%s_receiver, seg, 0, %s_sender, 1, DMAPP_QW);\n", v, v);
                        os.println("\t\t" + DmappVisitor.errorCheck);
                    }
                }
                os.printf("\t}\n");
            }

        }

        os.printf("\tdmapp_gsync_wait();\n");
        os.printf("\tPMI_Barrier();\n");

        /* check outputs */

        os.printf("\n\t/* Load results from global _received variables to locals */\n");
        os.printf("\tif %s {\n", checkThread(instanceValue.processes.size(), 0));
        for (Proc p : instanceValue.processes) {
            for (Var v : varsValue.procToLocal.get(p)) {
                os.printf("\t\t%s = *%s_receiver;\n", v, v);
            }
        }
        os.printf("\t}\n");

        for (Proc p : instanceValue.processes) {
            for (Var v : varsValue.procToLocal.get(p)) {
                if (firstVar) {
                    firstVar = false;
                } else {
                    varFormats.append(" ");
                }
                varNames.append(", ");

                String vf = String.format("%s=%%lld", v );
                varFormats.append(vf);
                varNames.append(v);
            }
        }

        os.printf("\n\t/* Check which output is observed */\n");

        os.printf("\tif %s {\n", checkThread(instanceValue.processes.size(), 0));
	if (instanceValue.outputs.size() == 0)
	    os.println("\t\tif (0) ;");
        for (int i = 0; i < instanceValue.outputs.size(); i++) {
            String output = instanceValue.outputs.get(i);
            os.printf("\t\t%sif %s cnt_%d++;\n", (i == 0)? "" : "else ",  output, instanceValue.outputs.indexOf(output));
        }
        os.printf("\t\telse fprintf(stderr, \"UNEXPECTED: %s\\n\"%s);\n", varFormats.toString(), varNames.toString());

        os.printf("\t}\n");
    }

    public void incorporateVarReleases(PrintStream os) throws java.io.IOException {
        /*free the shared variables */

        for (Proc p : varsValue.procToShared.keySet()) {
            for (Var v : varsValue.procToShared.get(p)) {
                os.printf("\tdmapp_sheap_free(%s);\n", v.name);
            }
        }


    }

    public void incorporateStats(PrintStream os) throws java.io.IOException {
        /* print the execution outputs */

        os.printf("\tif %s {\n", checkThread(instanceValue.processes.size(), 0));
        for (int i = 0; i < instanceValue.outputs.size(); i++) {
            String output = instanceValue.outputs.get(i);
            os.println("\t\tfprintf(stderr, \"OBSERVED " + output + ": %ld\\n\", cnt_" + instanceValue.outputs.indexOf(output) + ");");
        }
        os.printf("\t}\n");
    }

    public void macroExpand(PrintStream os, int payload) throws java.io.IOException {
        switch (payload) {
        case 0:
            incorporateOriginalIR(os);
            break;
        case 1:
            incorporateVarDecls(os);
            break;
        case 2:
            incorporateVars(os);
            break;
        case 3:
            incorporateTestBody(os);
            break;
        case 4:
            incorporateTestWitnesses(os);
            break;
        case 5:
            incorporateVarReleases(os);
            break;
        case 6:
            incorporateStats(os);
            break;
        default:
            os.printf("$%d\n", payload);
            break;
        }
    }

    public void expandTemplate(PrintStream out) throws java.io.IOException {
        /* Get template */
        InputStream in = ConvertToDmapp.class.getClassLoader().getResourceAsStream("dmapp_template.c");
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        String s;

        while ((s = br.readLine()) != null) {
            if (s.trim().startsWith("$")) {
                String st = s.trim();
                if (st.endsWith(";")) st = st.substring(0, st.length()-1);
                int payload = Integer.parseInt(st.substring(1));
                macroExpand(out, payload);
            } else {
                out.println(s);
            }
        }

        br.close();
    }

    public void generateScript(PrintStream out, String outputName, int numProcs) throws java.io.IOException {

        InputStream in = ConvertToDmapp.class.getClassLoader().getResourceAsStream("script_template.sbatch");
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        String s, sNumProcs;

        sNumProcs = Integer.toString(numProcs);
        while ((s = br.readLine()) != null) {
            String st = s.trim();
            if (st.contains("$1")) {
                st = st.replace("$1", outputName);
            }
            if (st.contains("$2")) {
                st = st.replace("$2", sNumProcs);
            }
            out.println(st);
        }
    }

    public static void main(String[] args) throws Exception {
        if (args.length < 2) {
            System.out.println("Please specify input file name and output name");
            System.exit(1);
        }
        String inputName = args[0];
        String outputName = args[1];

        File input = new File(inputName);
        if (!input.exists()) {
            System.out.println("Input does not exist");
            System.exit(1);
        }
        if (input.isFile()) {

            generateFiles(inputName, outputName, DEFAULT_OUTPUT_DIR);

        } else {

            assert (input.isDirectory());
            File outputDir = new File(outputName);


            if (!outputDir.exists()) {
                outputDir.mkdir();
                if (!outputDir.exists()) {
                    System.err.printf("could not create output directory %s; aborting.\n",
                            outputName);
                    System.exit(1);
                }
            }
            if (!outputDir.isDirectory()) {
                System.err.printf("output directory %s is not directory; aborting.\n",
                        outputName);
                System.exit(1);
            }


            File[]files = input.listFiles();

            for(File file: files) {
                String outName = file.getName();
                outName = outName.substring(0, outName.lastIndexOf('.'));
                generateFiles(file.getAbsolutePath(), outName, outputName);
            }
        }

    }

    public static void generateFiles(String inputFileName, String outputFileName, String outputDirName) throws Exception {
        try {

            /* Parse input program */
            TLLexer lexer = new TLLexer(new ANTLRFileStream(inputFileName));
            TLParser parser = new TLParser(new CommonTokenStream(lexer));

            parser.setBuildParseTree(true);
            ParseTree tree = parser.parse();

            ConvertToDmapp c = new ConvertToDmapp(inputFileName, tree);

            File outputDir = new File(outputDirName);
            if (!outputDir.exists()) {
                outputDir.mkdir();
                if (!outputDir.exists()) {
                    System.err.printf("could not create output directory %s; aborting.\n",
                            outputDirName);
                    System.exit(1);
                }
            }
            if (!outputDir.isDirectory()) {
                System.err.printf("output directory %s is not directory; aborting.\n",
                        outputDirName);
                System.exit(1);
            }

            String codeFileName = outputFileName + ".c";
            File oc = new File(outputDir, codeFileName);
            PrintStream pc = new PrintStream(oc);

            c.expandTemplate(pc);

            pc.close();

            String scriptFileName = outputFileName + ".sbatch";
            File os = new File(outputDir, scriptFileName);
            PrintStream ps = new PrintStream(os);

            c.generateScript(ps, outputFileName, c.instanceValue.processes.size());
            ps.close();

        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
