// -*-  indent-tabs-mode:nil; c-basic-offset:4; -*-
import java.io.*;

import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;

import org.antlr.v4.runtime.ANTLRFileStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTree;

public class ConvertToPortals { // VPI_Verbs API
    public static String DEFAULT_OUTPUT_DIR = "output";

    String inputFileName;
    ParseTree tree;
    VarsValue varsValue;
    InstanceValue instanceValue;
    TranslateValue translateValue;

    public ConvertToPortals(String inputFileName, ParseTree tree) {
        this.inputFileName = inputFileName;
        this.tree = tree;

        translateValue = new TranslateValue();
        TranslateVisitor translateVisitor = new TranslateVisitor(translateValue);
        translateVisitor.visit(tree);

        instanceValue = new InstanceValue();
        FirstPortalsVisitor firstPortalsVisitor = new FirstPortalsVisitor(instanceValue);
        firstPortalsVisitor.visit(tree);

        varsValue = new VarsValue();
        VarsVisitor varsVisitor = new VarsVisitor(varsValue, instanceValue);
        varsVisitor.visit(tree);

        SecondPortalsVisitor secondPortalsVisitor = new SecondPortalsVisitor(varsValue, instanceValue);
        secondPortalsVisitor.visit(tree);
    }

    public void incorporateOriginalIR(PrintStream os) throws java.io.IOException {
        BufferedReader br = new BufferedReader(new FileReader(inputFileName));
        String s = null;
        while ((s = br.readLine()) != null) {
            os.println(s);
        }
        br.close();
    }

    public void incorporateSharedVars(PrintStream os) throws java.io.IOException {
        Set<Var> allVars = new HashSet<>();
        for (Proc p : instanceValue.processes) {
            allVars.addAll(varsValue.procToShared.get(p));
        }

        for (Var v : allVars) {
            os.printf("    uint64_t *      %s = &vars[%d];\n", v.name, instanceValue.varNameToIndex.get(v.name));
        }
    }
    
    public void incorporateLocals(PrintStream os) throws java.io.IOException {
        Set<Var> localVars = new HashSet<>();

        StringBuilder vars = new StringBuilder();
        for (Proc p : instanceValue.processes) {
            localVars.addAll(varsValue.procToLocal.get(p));
        }

        for (Var v : localVars) {
            if (vars.length() > 0) vars.append(", ");
            vars.append(v.name);
        }

        /* add the counter variables, for each possible outcome */
        for (int i = 0; i < instanceValue.outputs.size(); i++) {
        	if (vars.length() > 0) vars.append(", ");
            vars.append("output_cnt_" + i + " = 0");
        }

        os.printf("    uint64_t        %s;\n", vars);
    }

    public void incorporateVarInitializations(PrintStream os) throws java.io.IOException {
        LinkedList<Var> allVars = new LinkedList<>();

        for (Proc p : instanceValue.processes) {
            for (Var v : varsValue.procToShared.get(p)) {
                allVars.add(v);
            }
        } 

        for (Proc p : instanceValue.processes) {
            os.printf("        if (%d == rank) {\n", p.procNumber);

            List<Var> otherProcVars = new ArrayList<>();
            otherProcVars.addAll(allVars);
            otherProcVars.removeAll(varsValue.procToShared.get(p));

            // initialization of var values
            for (Var v : varsValue.procToShared.get(p)) {
                if (v.hasInitialValue)
                    os.printf("            *%s = %d;\n", v.name, v.iv);
                else
                    os.printf("            *%s = 0;\n", v.name);
            }

            for (Var v : otherProcVars) {
                os.printf("            *%s = 0;\n", v.name);
            }

            for (Var v : varsValue.procToLocal.get(p)) {
                os.printf("            // initializing local variable: %s\n", v.name);
                os.printf("            %s = 0;\n", v.name);
            }

            os.printf("        }\n");
        }
    }
        
    
    public void incorporateTestBody(PrintStream os) throws java.io.IOException {
        for (Proc p : instanceValue.processes) {
            os.printf("        if (%d == rank) {\n", p.procNumber);
            if (p.procNumber == 0) {
                /* add long sleep for first thread */
                os.println("            if (rand() % 2 == 1) { usleep(5 + (rand()%10)); }");
            }
            os.printf("            ptl_process_t this, peer;\n");
            os.printf("            this.rank = %d;\n", p.procNumber);
            os.printf("            peer.rank = %d;\n", 1-p.procNumber);
            for (String s : instanceValue.processContents.get(p.procNumber)) {
                if (!s.contains("ctc.failure")) {
                    os.println("            if (rand() % 2 == 1) { usleep(1); }");
                }
                os.println("            " + s);
            }

            if (instanceValue.mdEventCounts.get(p.procNumber) > 0) {
                os.printf("            CHECK_RETURNVAL(PtlCTWait(vars_md.ct_handle, %d, &ctc));\n", instanceValue.mdEventCounts.get(p.procNumber));
                os.printf("            assert(ctc.failure == 0);\n");
            }
            if (instanceValue.leEventCounts.get(p.procNumber) > 0) {
                os.printf("            CHECK_RETURNVAL(PtlCTWait(vars_le.ct_handle, %d, &ctc));\n", instanceValue.leEventCounts.get(p.procNumber));
                os.printf("            assert(ctc.failure == 0);\n");
            }
            os.println();
            os.printf("        }\n");
        }
    }

    public void incorporateTestAsserts(PrintStream os) throws java.io.IOException {
        
        os.printf("        /* CHECKING OUTPUT */ \n");
        
        /* Put local variables in corresponding _sender shared variables. */
        for (Proc p : instanceValue.processes) {
            if (p.procNumber != 0) {
                os.printf("        if (rank == %d) {\n", p.procNumber);
                os.printf("            ptl_process_t peer;\n");
                os.printf("            peer.rank = 0;\n");
                for (Var v: varsValue.procToLocal.get(p)) {
                    /* For all processes except Process 0, send _sender variables to corresponding _receiver variable on Process 0 */
                    os.printf("            *%s_sender = %s;\n", v, v);
                    os.printf("            // put(%s, %d, %s)\n", v + "_receiver", 0, v + "_sender");
                    os.printf
                        ("            CHECK_RETURNVAL(PtlPut(vars_md_handle, %d*sizeof(uint64_t), sizeof(uint64_t), PTL_CT_ACK_REQ, peer, pt_index, 1, %d*sizeof(uint64_t), NULL, 0));\n",
                                                instanceValue.varNameToIndex.get(v + "_sender"), instanceValue.varNameToIndex.get(v + "_receiver"));
                        }
                if (varsValue.procToLocal.get(p).size() > 0) {
                    os.printf("            CHECK_RETURNVAL(PtlCTWait(vars_md.ct_handle, 2*%d, &ctc));\n", varsValue.procToLocal.get(p).size()); // again, 2 events per put
                    os.printf("            assert(ctc.failure == 0);\n");
                }
                os.printf("        }\n");    
        	}
        }
        
        os.printf("        BARRIER;\n");
        
        /* Load all _receiver shared variables in the local variables */
        os.printf("        if (rank == 0) {\n");
        for (Proc p : instanceValue.processes) {
            if (p.procNumber != 0) {
                for (Var v: varsValue.procToLocal.get(p)) {
                    os.printf("            %s = *%s_receiver;\n", v, v);
                }
            }
        }
        
        /* Increment the counter corresponding to the observed output */
        for(int i = 0; i < instanceValue.outputs.size(); i++) {
            String output = instanceValue.outputs.get(i);
            if (i == 0) {
                os.printf("            if (%s) output_cnt_%d += 1;\n", output, i);
            } else {
                os.printf("            else if (%s) output_cnt_%d += 1;\n", output, i);
            }
        }
        
        
        // print out: all locals for rank 0
        StringBuilder rank0LocalVarFormats = new StringBuilder(),
            rank0LocalVarNames = new StringBuilder();

        // generate local variable names
        for (Proc p : instanceValue.processes) {
            boolean firstVar = true;
            for (Var lv : varsValue.procToLocal.get(p)) {
                if (firstVar) {
                    firstVar = false;
                } else {
                    rank0LocalVarFormats.append(" ");
                }
                rank0LocalVarNames.append(", ");

                String vf = String.format("%s=%%lld ", lv);
                rank0LocalVarFormats.append(vf);
                rank0LocalVarNames.append(lv);
            }
        }
        
        os.printf("            else printf(\" FAILED. Output: %s\\n\"%s);\n",
                rank0LocalVarFormats.toString(), rank0LocalVarNames.toString());
        
        os.printf("        }\n");  
    }
    
    private void incorporatePrintStats(PrintStream os) {
        os.printf("    if (rank == 0) {\n");
        for(int i = 0; i < instanceValue.outputs.size(); i++) {
            String output = instanceValue.outputs.get(i);
            os.println("\tprintf(\"OBSERVED " + output + ": %ld\\n\", output_cnt_" + instanceValue.outputs.indexOf(output) + ");");
        }
        os.printf("    }\n");
    }

    
    StringBuilder simpleExpressionBuffer;
    
    private Object visitSimpleExpression(TLParser.SimpleExpressionContext ctx) {
        if (ctx.And() != null) {
            /* simpleExpression And simpleExpression*/
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" && ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Or() != null) {
            /* simpleExpression Or simpleExpression */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" || ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Excl() != null) {
            /* Excl simpleExpression */
            simpleExpressionBuffer.append("!(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        } else if (ctx.OParen() != null) {
            /* OParen simpleExpression CParen */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Equals() != null) {
            /* Identifier Equals rhs */
            simpleExpressionBuffer.append(ctx.Identifier().getText() + " == " );
            visitRhs(ctx.rhs());

        } else if (ctx.NEquals() != null) {
            /* Identifier NEquals rhs */
            simpleExpressionBuffer.append("(" + ctx.Identifier().getText() + " != " );
            visitRhs(ctx.rhs());
            simpleExpressionBuffer.append(")");

        } else {
            System.err.println("Error");
            System.exit(1);
        }
        return null;
    }

    public Object visitRhs(TLParser.RhsContext ctx) {
        if (ctx.Number() != null) {
            simpleExpressionBuffer.append(ctx.getText()+"L");
        } else {
            simpleExpressionBuffer.append(ctx.getText());
        }
        return null;
    }

    public void macroExpand(PrintStream os, int payload, String outputName) throws java.io.IOException {
        switch (payload) {
        case 0:
            os.printf("%s\n", outputName);
            break;
        case 1:
            incorporateOriginalIR(os);
            break;

        case 2:
            incorporateSharedVars(os);
            break;
        case 3:
            incorporateLocals(os);
            break;
        case 4:
            incorporateVarInitializations(os);
            break;

        case 5:
            incorporateTestBody(os);
            break;
        case 6:
            incorporateTestAsserts(os);
            break;
            
        case 7:
            incorporatePrintStats(os);
            break;

        default:
            os.printf("$%d\n", payload);
            break;
        }
    }

    

    public void expandTemplate(File outputDir, String outputName) throws java.io.IOException {
        String fn = "portals_test_generated.c";
        File os = new File(outputDir, fn);
        PrintStream ps = new PrintStream(os);

        InputStream in = ConvertToPortals.class.getClassLoader().getResourceAsStream(fn);
        BufferedReader br = new BufferedReader(new InputStreamReader(in));
        String s;

        while ((s = br.readLine()) != null) {
            String st = s.trim();
            if (st.contains("{0}")) {
                st = s.replace("{0}", outputName);
                ps.println(st);
            } else if (st.contains("{1}")) {
                st = s.replace("{1}", Integer.toString(instanceValue.processes.size()));
                ps.println(st);
            } else if (st.contains("{2}")) {
                int biggestIndex = -1;
                for (String v: instanceValue.varNameToIndex.keySet()) {
                    if (instanceValue.varNameToIndex.get(v) > biggestIndex)
                        biggestIndex = instanceValue.varNameToIndex.get(v);
                }
                st = s.replace("{2}", Integer.toString(biggestIndex+1));
                ps.println(st);
            } else if (st.startsWith("$")) {
                if (st.endsWith(";")) st = st.substring(0, st.length()-1);
                int payload = Integer.parseInt(st.substring(1));
                macroExpand(ps, payload, outputName);
            } else {
                ps.println(s);
            }
        }
        br.close();
        ps.close();
    }

    public static void main(String[] args) throws Exception {
        if (args.length < 2) {
            System.out.println("Please specify input file name and output name");
            System.exit(1);
        }
        String inputFileName = args[0];
        String outputFileName = args[1];

        /* Parse input program */
        TLLexer lexer = new TLLexer(new ANTLRFileStream(inputFileName));
        TLParser parser = new TLParser(new CommonTokenStream(lexer));

        parser.setBuildParseTree(true);
        ParseTree tree = parser.parse();

        ConvertToPortals c = new ConvertToPortals(inputFileName, tree);

        String allOutputDirName = DEFAULT_OUTPUT_DIR;
        File allOutputDir = new File(allOutputDirName);
        if (!allOutputDir.exists()) {
            allOutputDir.mkdir();
            if (!allOutputDir.exists()) {
                System.err.printf("could not create output directory %s; aborting.\n",
                                  allOutputDirName);
                System.exit(1);
            }
        }
        if (!allOutputDir.isDirectory()) {
            System.err.printf("output directory %s is not directory; aborting.\n",
                              allOutputDirName);
            System.exit(1);
        }

        String thisOutputDirName = allOutputDir + File.separator + outputFileName;
        File thisOutputDir = new File(thisOutputDirName);
        if (!thisOutputDir.exists()) {
            thisOutputDir.mkdir();
            if (!thisOutputDir.exists()) {
                System.err.printf("could not create output directory %s; aborting.\n",
                                  thisOutputDirName);
                System.exit(1);
            }
        }
        if (!thisOutputDir.isDirectory()) {
            System.err.printf("output directory %s is not directory; aborting.\n",
                              thisOutputDirName);
            System.exit(1);
        }

        c.expandTemplate(thisOutputDir, outputFileName);
    }
}
