// -*-  indent-tabs-mode:nil; c-basic-offset:4; -*-
// current assumption: at most 2 processes

import java.util.List;
import java.util.ArrayList;

public class SecondPortalsVisitor extends TLBaseVisitor<Object> {
    VarsValue vars;
    InstanceValue result;
    StringBuilder simpleExpressionBuffer, expressionBuffer;

    Proc currentProc;
    int counter;

    public SecondPortalsVisitor(VarsValue vars, InstanceValue result) {
        this.vars = vars;
        this.result = result;
        simpleExpressionBuffer = new StringBuilder();
        expressionBuffer = new StringBuilder();
    }

    @Override
    public Object visitProcess(TLParser.ProcessContext ctx) {
        int procNumber = Integer.parseInt(ctx.Number().getText());
        for (Proc p : result.processes) {
            if (p.procNumber == procNumber)
                this.currentProc = p;
        }

        visitChildren(ctx);
        return null;
    }

    @Override
    public Object visitSharedDecl(TLParser.SharedDeclContext ctx) {
        return null;
    }

    @Override
    public Object visitLocalDecl(TLParser.LocalDeclContext ctx) {
        return null;
    }

    @Override
    public Object visitStatement(TLParser.StatementContext ctx) {
        visitChildren(ctx);
        return null;
    }

    @Override
    public Object visitAssignment(TLParser.AssignmentContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String lhs = ctx.Identifier().getText();

        expressionBuffer.setLength(0);
        if (ctx.expression() instanceof TLParser.NumberExpressionContext)
            visitNumberExpression((TLParser.NumberExpressionContext)ctx.expression());
        else {
            System.out.println("unhandled expr in assignment stmt");
            System.exit(1);
        }

        String rhs = expressionBuffer.toString();

        String newInst = String.format("%s = %s;", lhs, rhs);
        insts.add(newInst);

        return null;
    }

    @Override
    public Object visitFlush(TLParser.FlushContext ctx) {
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        int thisProcessNumber = this.currentProc.procNumber;
        List<String> insts = result.processContents.get(thisProcessNumber);
        String inst = String.format("CHECK_RETURNVAL(PtlCTWait(vars_md.ct_handle, %d, &ctc));",
                                    result.mdEventCounts.get(thisProcessNumber));
        insts.add(inst);
        result.mdEventCounts.put(thisProcessNumber, 0);
        insts.add("assert(ctc.failure == 0);");

        visitChildren(ctx);
        return null;
    }

    @Override
    public Object visitStore(TLParser.StoreContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String writeVar = ctx.Identifier().getText();
        String rhs = ctx.rhs().getText();
        String writeAtomicity = ctx.atomicity().getText();

        String newInst = String.format("*%s = %s;", writeVar, rhs);
        insts.add(newInst);

        return null;
    }

    @Override
    public Object visitLoad(TLParser.LoadContext ctx) {
        List<String> insts = result.processContents.get(this.currentProc.procNumber);

        String lhs = ctx.Identifier(0).getText();
        String readVar = ctx.Identifier(1).getText();
        String readAtomicity = ctx.atomicity().getText();

        String newInst = String.format("%s = *%s;", lhs, readVar);
        insts.add(newInst);

        return null;
    }

    @Override
    public Object visitGet(TLParser.GetContext ctx) {
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        int thisProcessNumber = this.currentProc.procNumber;
        List<String> insts = result.processContents.get(thisProcessNumber);

        String writeVar = ctx.Identifier(0).getText();
        String readVar = ctx.Identifier(1).getText();
        String readName = "vr" + counter;
        String writeName = "vw" + counter;

        String writeAtomicity = ctx.atomicity(0).getText();
        String readAtomicity = ctx.atomicity(1).getText();

        String dest = null;
        if (thisProcessNumber == destProcessNumber) {
            dest = "this";
        } else {
            dest = "peer";
        }

        insts.add(String.format("// %s = get(%s, %d)", writeVar, readVar, destProcessNumber));
        insts.add(String.format("CHECK_RETURNVAL(PtlGet(vars_md_handle, %d*sizeof(uint64_t), sizeof(uint64_t), %s, pt_index, 1, %d*sizeof(uint64_t), NULL));", result.varNameToIndex.get(writeVar), dest, result.varNameToIndex.get(readVar)));
        
        int srcExpectedEventCount = result.mdEventCounts.get(thisProcessNumber);
        srcExpectedEventCount += 1; // portals spec page 92: 1 src event for a get
        result.mdEventCounts.put(thisProcessNumber, srcExpectedEventCount);

        int destExpectedEventCount = result.leEventCounts.get(destProcessNumber);
        destExpectedEventCount += 1;
        result.leEventCounts.put(destProcessNumber, destExpectedEventCount);

        return null;
    }

    @Override
    public Object visitPut(TLParser.PutContext ctx) {
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        int thisProcessNumber = this.currentProc.procNumber;
        List<String> insts = result.processContents.get(thisProcessNumber);

        String writeVar = ctx.Identifier(0).getText();
        String readVar = ctx.Identifier(1).getText();
        String readName = "vr" + counter;
        String writeName = "vw" + counter;

        String writeAtomicity = ctx.atomicity(0).getText();
        String readAtomicity = ctx.atomicity(1).getText();

        String dest = null;
        if (thisProcessNumber == destProcessNumber) {
            dest = "this";
        } else {
            dest = "peer";
        }

        insts.add(String.format("// put(%s, %d, %s)", writeVar, destProcessNumber, readVar));
        String newInst = String.format
            ("CHECK_RETURNVAL(PtlPut(vars_md_handle, %d*sizeof(uint64_t), sizeof(uint64_t), PTL_CT_ACK_REQ, %s, pt_index, 1, %d*sizeof(uint64_t), NULL, 0));",
                        result.varNameToIndex.get(readVar), dest, result.varNameToIndex.get(writeVar));
        insts.add(newInst);

        int srcExpectedEventCount = result.mdEventCounts.get(thisProcessNumber);
        srcExpectedEventCount += 2; // portals spec page 89: 2 events for a put (2 src, 1 dest)
        result.mdEventCounts.put(thisProcessNumber, srcExpectedEventCount);

        int destExpectedEventCount = result.leEventCounts.get(destProcessNumber);
        destExpectedEventCount += 1;
        result.leEventCounts.put(destProcessNumber, destExpectedEventCount);

        return null;
    }

    @Override
    public Object visitRga(TLParser.RgaContext ctx) {
        // X = rga(D, d, Y)
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        int thisProcessNumber = this.currentProc.procNumber;
        List<String> insts = result.processContents.get(thisProcessNumber);

        String XVar = ctx.Identifier(0).getText(); // local-write
        String DVar = ctx.Identifier(1).getText(); // remote-write
        String YVar = ctx.Identifier(2).getText(); // read

        String dest = null;
        if (thisProcessNumber == destProcessNumber) {
            dest = "this";
        } else {
            dest = "peer";
        }

        insts.add(String.format("// %s = rga(%s, %d, %s)", XVar, DVar, destProcessNumber, YVar));
        String newInst = String.format
            ("CHECK_RETURNVAL(PtlFetchAtomic(vars_md_handle, %d*sizeof(uint64_t), vars_md_handle, %d*sizeof(uint64_t), sizeof(uint64_t), %s, pt_index, 1, %d*sizeof(uint64_t), NULL, 0, PTL_SUM, PTL_UINT64_T));", result.varNameToIndex.get(YVar), result.varNameToIndex.get(XVar), dest, result.varNameToIndex.get(DVar));
        insts.add(newInst);

        int srcExpectedEventCount = result.mdEventCounts.get(thisProcessNumber);
        srcExpectedEventCount += 2; // portals spec page 97: 2 events for a rga
        result.mdEventCounts.put(thisProcessNumber, srcExpectedEventCount);

        int destExpectedEventCount = result.leEventCounts.get(destProcessNumber);
        destExpectedEventCount += 1;
        result.leEventCounts.put(destProcessNumber, destExpectedEventCount);

        return null;
    }

    @Override
    public Object visitCas(TLParser.CasContext ctx) {
        // X = cas(D, d, Y, W)
        int destProcessNumber = Integer.parseInt(ctx.Number().getText());
        int thisProcessNumber = this.currentProc.procNumber;
        List<String> insts = result.processContents.get(thisProcessNumber);

        String XVar = ctx.Identifier(0).getText(); // w
        String DVar = ctx.Identifier(1).getText(); // rw
        String YVar = ctx.Identifier(2).getText(); // r
        String WVar = ctx.Identifier(3).getText(); // r

        String dest = null;
        if (thisProcessNumber == destProcessNumber) {
            dest = "this";
        } else {
            dest = "peer";
        }

        insts.add(String.format("// %s = cas(%s, %d, %s, %s)", XVar, DVar, destProcessNumber, YVar, WVar));
        String newInst = String.format
            ("CHECK_RETURNVAL(PtlSwap(vars_md_handle, %d*sizeof(uint64_t), vars_md_handle, %d*sizeof(uint64_t), sizeof(uint64_t), %s, pt_index, 1, %d*sizeof(uint64_t), NULL, 0, %s, PTL_CSWAP, PTL_UINT64_T));", result.varNameToIndex.get(XVar), result.varNameToIndex.get(YVar), dest, result.varNameToIndex.get(DVar), WVar);
        insts.add(newInst);

        int srcExpectedEventCount = result.mdEventCounts.get(thisProcessNumber);
        srcExpectedEventCount += 1; // portals spec page 97: 1 event for a cas
        result.mdEventCounts.put(thisProcessNumber, srcExpectedEventCount);

        int destExpectedEventCount = result.leEventCounts.get(destProcessNumber);
        destExpectedEventCount += 1;
        result.leEventCounts.put(destProcessNumber, destExpectedEventCount);

        return null;
    }

    public Object visitRhs(TLParser.RhsContext ctx) {
        if (ctx.Number() != null) {
            simpleExpressionBuffer.append(ctx.getText());
        } else {
            simpleExpressionBuffer.append("*("+ctx.getText() + ")");
        }
        return null;
    }

    @Override
    public Object visitNumberExpression(TLParser.NumberExpressionContext ctx) {
	expressionBuffer.append(ctx.Number());
	return null;
    }

    public Object visitSimpleExpression(TLParser.SimpleExpressionContext ctx) {
        if (ctx.And() != null) {
            /* simpleExpression And simpleExpression*/
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" && ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Or() != null) {
            /* simpleExpression Or simpleExpression */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(" || ");
            visitSimpleExpression(ctx.simpleExpression(1));
            simpleExpressionBuffer.append(")");

        } else if (ctx.Excl() != null) {
            /* Excl simpleExpression */
            simpleExpressionBuffer.append("! (");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        } else if (ctx.OParen() != null) {
            /* OParen simpleExpression CParen */
            simpleExpressionBuffer.append("(");
            visitSimpleExpression(ctx.simpleExpression(0));
            simpleExpressionBuffer.append(")");

        }else if (ctx.Equals() != null) {
            /* Identifier Equals rhs */
            simpleExpressionBuffer.append("" + ctx.Identifier().getText() + " == " );
            visitRhs(ctx.rhs());

        } else if (ctx.NEquals() != null) {
            /* Identifier NEquals rhs */
            simpleExpressionBuffer.append("(*(" + ctx.Identifier().getText() + ") != " );
            visitRhs(ctx.rhs());
            simpleExpressionBuffer.append(")");

        } else {
            System.err.println("Error");
            System.exit(1);
        }
        return null;
    }

    public Object visitAssumption(TLParser.AssumptionContext ctx) {
        /* Reset the buffer */
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        // System.out.println("Assumption found: " + simpleExpressionBuffer.toString());

        return null;
    }

    public Object visitAssertion(TLParser.AssertionContext ctx) {
        /* Reset the buffer */
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        /* Append buffer to result */
        result.owningProcess.put(ctx.simpleExpression(), currentProc.procNumber);
        result.assertions.add(ctx.simpleExpression());

        // System.out.println("Assertion found: " + result.Assertion.toString());

        return null;
    }

    public Object visitOutput(TLParser.OutputContext ctx) {
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        result.outputs.add(simpleExpressionBuffer.toString());
        return null;
    }

    public Object visitIfStatement(TLParser.IfStatementContext ctx) {
        simpleExpressionBuffer.setLength(0);

        visitSimpleExpression(ctx.simpleExpression());

        return null;
    }

}
