#include <stdio.h>
#include <unistd.h>
#include "pmi.h"
#include "dmapp.h"

/*
	T1:			T2:			T3:			T4:
	X = 0;			A1 = get(Y)		B1 = get(Y)
	put(Y, X)		A2 = get(Y)		B2 = get(Y)
	Z = 1;
	put(Y, Z)				

	spec: it cannot be (A1 = B2 = 0 and A2 = B1 = 1) or (A1 = B2 = 1 and A2 = B1 = 0)

*/
int main(int argc, char **argv) {
	char 			*expl = NULL;
	int 			pe = -1;
	long 			*X = NULL;	
	long 			*Y = NULL;
	long			*Z = NULL;
	long			*A = NULL;
	long 			*B = NULL;
	dmapp_return_t 		status;
	dmapp_rma_attrs_t 	actual_args = {0}, rma_args = {0};
	dmapp_jobinfo_t 	job;
	dmapp_seg_desc_t 	*seg = NULL;
	
	/* Set the RMA parameters. */
	rma_args.put_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
	rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
	rma_args.max_concurrency = 1;
	
	/* Initialize DMAPP resources before executing any other DMAPP calls. */
	status = dmapp_init(&rma_args, &actual_args);
	if(status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
		exit(1);
	}
	
	/* Retrieve information about job details such as PE id. */
	status = dmapp_get_jobinfo(&job);
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	
	pe = job.pe;

	/* Specify in which segment the remote memory region (the source) lies. 
	   In this case, it is the heap (see above). */
  	seg = &(job.sheap_seg);
	
	/* Allocate remotely accessible memory for the source and target bufferes. 
	   Only memory in the data segment or the sheap is remotely accessible. 
	   Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	Z = (long *)dmapp_sheap_malloc(sizeof(long));
	A = (long *)dmapp_sheap_malloc(sizeof(long));
	B = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL || Y == NULL || Z == NULL || A == NULL || B == NULL) {
		fprintf(stderr, " sheap_malloc'd failed X 0x%lx Y 0x%lx Z 0x%lx A 0x%lx B 0x%lx\n", (long)X, (long)Y, (long)Z, (long)A, (long)B);	
		exit(1);
	}
	
	/* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */ 
	PMI_Barrier();	
	
	/* Initialization. */
	if(pe == 0) {
		*X = 0L;
		status = dmapp_put_nbi(Y, seg, 3, X, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_put FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
		*Z = 1L;
		status = dmapp_put_nbi(Y, seg, 3, Z, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_put FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
	} else if (pe == 1) {
		status = dmapp_get_nbi(X, Y, seg, 3, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
		status = dmapp_get_nbi(Y, Y, seg, 3, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
	} else if (pe == 2) {
		status = dmapp_get_nbi(X, Y, seg, 3, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
                status = dmapp_get_nbi(Y, Y, seg, 3, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
	} else if (pe == 3) {
	} else {
		fprintf(stderr, "Unhandled behaviour!");
		exit(1);
	}
	
	/* Synchronize before verifying the data. */
	PMI_Barrier();

	/* Verify the data. */
	if (pe == 1) {
		status = dmapp_get_nbi(A, Y, seg, 2, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
                status = dmapp_get_nbi(B, X, seg, 2, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
		dmapp_gsync_wait();
		if((*X == 0L && *A == 0L && *Y == 1L && *B == 1L) || (*X == 1L && *A == 1L && *Y == 0L && *B == 0L)) {
			fprintf(stderr, " *PE_1_FAILED(A1=%ld;A2=%ld;B1=%ld;B2=%ld;)* ", *X, *Y, *B, *A);
		} else {
			fprintf(stderr, " *PE_1_SUCCESS(A1=%ld;A2=%ld;B1=%ld;B2=%ld;)* ", *X, *Y, *B, *A);
		}
	} else if (pe == 2) {
		status = dmapp_get_nbi(A, Y, seg, 1, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
                status = dmapp_get_nbi(B, X, seg, 1, 1, DMAPP_QW);
                if (status != DMAPP_RC_SUCCESS) {
                        dmapp_explain_error(status, (const char **)&expl);
                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                        exit(1);
                }
                dmapp_gsync_wait();
                if((*X == 0L && *A == 0L && *Y == 1L && *B == 1L) || (*X == 1L && *A == 1L && *Y == 0L && *B == 0L)) {
                        fprintf(stderr, " *PE_2_FAILED(A1=%ld;A2=%ld;B1=%ld;B2=%ld;)* ", *B, *A, *X, *Y);
                } else {
			fprintf(stderr, " *PE_2_SUCCESS(A1=%ld;A2=%ld;B1=%ld;B2=%ld;)* ", *B, *A, *X, *Y);
		}
	}
	
	/* Synchronize before clearing buffers. */
	PMI_Barrier();
	
	/* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(Y);
	dmapp_sheap_free(Z);	
	dmapp_sheap_free(A);
	dmapp_sheap_free(B);

	/* Release DMAPP resources. This is a mandatory call. */
	status = dmapp_finalize();
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	

	return(0);
}

