#include <stdio.h>
#include <unistd.h>
#include "pmi.h"
#include "dmapp.h"
#include <pthread.h>
#define NUM_THREADS    10

/*
	Stores are not reordered with other stores (x86-TSO).
	Loads are not reordered with older loads (x86-TSO).	
	
	P1:					P2:
	T1:			T2:		X = 0
	X = remoteGet(X)	a = Y
	flush			b = X
	X = 1			
	Y = 1			

	spec: (a == 1L && b == 1L) || (a == 0L && b == 1L) || (a == 0L && b == 0L)
*/

char                    *expl = NULL;
int                     pe = -1;
int                     npes = -1;
int                     target_pe = -1;
long                    *X = NULL;
long                    *Y = NULL;
dmapp_return_t          status;
dmapp_rma_attrs_t       actual_args = {0}, rma_args = {0};
dmapp_jobinfo_t         job;
dmapp_seg_desc_t        *seg = NULL;

void *t1(void *threadid) {
	status = dmapp_get_nbi(X, X, seg, target_pe, 1, DMAPP_QW);
        if (status != DMAPP_RC_SUCCESS) {
                dmapp_explain_error(status, (const char **)&expl);
                fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
        	exit(1);
	}
        dmapp_gsync_wait();
	*X = 1L;
	*Y = 1L;

	pthread_exit(NULL);	
}

void *t2(void *threadid) {
	long a = *Y;
	sleep(5);
	long b = *X;
	if(!((a == 1L && b == 1L) || (a == 0L && b == 1L) || (a == 0L && b == 0L))) {
		fprintf(stderr, "FAILED:(%ld,%ld);\n", a, b);
	}
	fprintf(stderr, "(%ld,%ld)", a, b);

	pthread_exit(NULL);
}

int main(int argc, char **argv) {
	/* Set the RMA parameters. */
	rma_args.put_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
	rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
	rma_args.max_concurrency = NUM_THREADS;
	
	/* Initialize DMAPP resources before executing any other DMAPP calls. */
	status = dmapp_init(&rma_args, &actual_args);
	if(status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
		exit(1);
	}
	
	/* Retrieve information about job details such as PE id and number of PEs. */
	status = dmapp_get_jobinfo(&job);
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	
	pe = job.pe;
	npes = job.npes;	
	target_pe = npes - pe - 1;

	/* Specify in which segment the remote memory region (the source) lies. 
	   In this case, it is the heap (see above). */
  	seg = &(job.sheap_seg);
	
	/* Allocate remotely accessible memory for the source and target bufferes. 
	   Only memory in the data segment or the sheap is remotely accessible. 
	   Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL || Y == NULL) {
		fprintf(stderr, " sheap_malloc'd failed X 0x%lx Y 0x%lx\n", (long)X, (long)Y);	
		exit(1);
	}
	
	/* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */ 
	PMI_Barrier();		
	/* Initialization (P1 == even PEs; P2 == odd PEs). */
	if((pe%2) == 0) {
		pthread_t threads[NUM_THREADS];
  		for(long t=0; t<NUM_THREADS; t=t+2) { 
			*X = 0L;
			*Y = 0L;
			if(t <= NUM_THREADS/2) {
				pthread_create(&threads[t], NULL, t1, (void *)t);
				pthread_create(&threads[t+1], NULL, t2, (void *)(t+1));
			} else {
				pthread_create(&threads[t], NULL, t2, (void *)t);
				pthread_create(&threads[t+1], NULL, t1, (void *)(t+1));
	                }
			sleep(1);
  		}		
	} else {
		*X = 0L;
	}
	
	/* Synchronize. */
	PMI_Barrier();
	
	/* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(Y);
	
	/* Release DMAPP resources. This is a mandatory call. */
	status = dmapp_finalize();
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	

	return(0);
}

