#include <stdio.h>
#include <unistd.h>
#include "pmi.h"
#include "dmapp.h"
#include <pthread.h>
#define NUM_THREADS     1000

/*
        Loads may be reordered with older stores (x86-TSO).
        
        P1:                                             P2:
        T1:                     T2:                     X = 1
        Y = remoteGet(Y)        X = remoteGet(X)        Y = 1                 
        flush                   flush
        X = 1                   Y = 1
        a = Y                   b = X                                                                        

        spec: (a == 1L && b == 0L) || (a == 0L && b == 1L) || (a == 0L && b == 0L) || (a == 1L && b == 1L)
*/

char                    *expl = NULL;
int                     pe = -1;
int                     npes = -1;
int                     target_pe = -1;
long                    *X = NULL;
long                    *Y = NULL;
dmapp_return_t          status;
dmapp_rma_attrs_t       actual_args = {0}, rma_args = {0};
dmapp_jobinfo_t         job;
dmapp_seg_desc_t        *seg = NULL;

void *t1(void *result) {
        status = dmapp_get_nbi(Y, Y, seg, target_pe, 1, DMAPP_QW);
        if (status != DMAPP_RC_SUCCESS) {
                dmapp_explain_error(status, (const char **)&expl);
                fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                exit(1);
        }
        dmapp_gsync_wait();
	*X = 1L;
	long a = *Y;
	
	*((long*)result) = a;
	pthread_exit(result);	
}

void *t2(void *result) {
        status = dmapp_get_nbi(X, X, seg, target_pe, 1, DMAPP_QW);
        if (status != DMAPP_RC_SUCCESS) {
                dmapp_explain_error(status, (const char **)&expl);
                fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
                exit(1);
        }
        dmapp_gsync_wait();

	*Y = 1L;
	long b = *X;
	
	*((long*)result) = b;
	pthread_exit(result);
}

int main(int argc, char **argv) {
	/* Set the RMA parameters. */
	rma_args.put_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
	rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
	rma_args.max_concurrency = 10;
	
	/* Initialize DMAPP resources before executing any other DMAPP calls. */
	status = dmapp_init(&rma_args, &actual_args);
	if(status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
		exit(1);
	}
	
	/* Retrieve information about job details such as PE id and number of PEs. */
	status = dmapp_get_jobinfo(&job);
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	
	pe = job.pe;
	npes = job.npes;	
	target_pe = npes - pe - 1;

	/* Specify in which segment the remote memory region (the source) lies. 
	   In this case, it is the heap (see above). */
  	seg = &(job.sheap_seg);
	
	/* Allocate remotely accessible memory for the source and target bufferes. 
	   Only memory in the data segment or the sheap is remotely accessible. 
	   Here we allocate from the sheap. */
	X = (long *)dmapp_sheap_malloc(sizeof(long));
	Y = (long *)dmapp_sheap_malloc(sizeof(long));
	if (X == NULL || Y == NULL) {
		fprintf(stderr, " sheap_malloc'd failed X 0x%lx Y 0x%lx\n", (long)X, (long)Y);	
		exit(1);
	}
	
	/* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */ 
	PMI_Barrier();	
	
	/* Initialization (P1 == even PEs; P2 == odd PEs). */
	if((pe%2) == 0) {
		pthread_t threads[NUM_THREADS];
  		for(long t=0; t<NUM_THREADS; t=t+2) { 
			*X = 0L;
			*Y = 0L;
			long a,b;
			if(t <= NUM_THREADS/2) {
				pthread_create(&threads[t], NULL, t1, &a);
				pthread_create(&threads[t+1], NULL, t2, &b);
			} else {
				pthread_create(&threads[t], NULL, t2, &b);
				pthread_create(&threads[t+1], NULL, t1, &a);
	                }
			sleep(1);

			void *status;
			pthread_join(threads[t], &status);
			pthread_join(threads[t+1], &status);
		
			if(!((a == 1L && b == 0L) || (a == 0L && b == 1L) || (a == 0L && b == 0L) || (a == 1L && b == 1L))) {
        		        fprintf(stderr, "FAILED:(%ld,%ld);\n", a, b);
	        	}
  		}		
	} else {
            	*X = 1L;
		*Y = 1L;
	}
	
	/* Synchronize. */
	PMI_Barrier();
	
	/* Free buffers allocated from sheap. */
	dmapp_sheap_free(X);
	dmapp_sheap_free(Y);
	
	/* Release DMAPP resources. This is a mandatory call. */
	status = dmapp_finalize();
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	

	return(0);
}

