#include <stdio.h>
#include <unistd.h>
#include "pmi.h"
#include "dmapp.h"
#include <stdlib.h>

#define NUM_ITERATIONS 10000000
#define ARRAY_SIZE 2

// Assumes 0 <= max <= RAND_MAX
// Returns in the half-open interval [0, max]
// sleep(random_at_most(10) / 10);
long random_at_most(long max) {
  unsigned long
    // max <= RAND_MAX < ULONG_MAX, so this is okay.
    num_bins = (unsigned long) max + 1,
    num_rand = (unsigned long) RAND_MAX + 1,
    bin_size = num_rand / num_bins,
    defect   = num_rand % num_bins;

  long x;
  do {
   x = random();
  }
  // This is carefully written not to overflow
  while (num_rand - defect <= (unsigned long)x);

  // Truncated division is intentional
  return x/bin_size;
}


int main(int argc, char **argv) {
	char 			*expl = NULL;
	int 			pe = -1;
	int 			npes = -1;
	int 			target_pe = -1;
	long long		*X = NULL;	
	long long		*Y = NULL;
	long long		*Z = NULL;
	dmapp_return_t 		status;
	dmapp_rma_attrs_t 	actual_args = {0}, rma_args = {0};
	dmapp_jobinfo_t 	job;
	dmapp_seg_desc_t 	*seg = NULL;
	
	/* Set the RMA parameters. */
	rma_args.put_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.get_relaxed_ordering = DMAPP_ROUTING_ADAPTIVE;
	rma_args.max_outstanding_nb = DMAPP_DEF_OUTSTANDING_NB;
	rma_args.offload_threshold = DMAPP_OFFLOAD_THRESHOLD;
	rma_args.max_concurrency = 1;
	
	/* Initialize DMAPP resources before executing any other DMAPP calls. */
	status = dmapp_init(&rma_args, &actual_args);
	if(status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_init returned %d - %s - %s \n", status, dmapp_err_str[status], expl);
		exit(1);
	}
	
	/* Retrieve information about job details such as PE id and number of PEs. */
	status = dmapp_get_jobinfo(&job);
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_get_jobinfo FAILED: w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	
	pe = job.pe;
	npes = job.npes;	
	target_pe = npes - pe - 1;

	/* Specify in which segment the remote memory region (the source) lies. 
	   In this case, it is the heap (see above). */
  	seg = &(job.sheap_seg);

	for(int i=0; i<NUM_ITERATIONS; i++) {
		/* Allocate remotely accessible memory for the source and target bufferes. 
		   Only memory in the data segment or the sheap is remotely accessible. 
		   Here we allocate from the sheap. */
		X = (long long *)dmapp_sheap_malloc(sizeof(long long) * ARRAY_SIZE);
		Y = (long long *)dmapp_sheap_malloc(sizeof(long long) * ARRAY_SIZE);
		Z = (long long *)dmapp_sheap_malloc(sizeof(long long) * ARRAY_SIZE);
		if (X == NULL || Y == NULL) {
			fprintf(stderr, " sheap_malloc'd failed X 0x%lx Y 0x%lx Z 0x%lx\n", (long)X, (long)Y, (long)Z);	
			exit(1);
		}
	
		if(pe%2 == 0) {
			*X = 9223372036854775807L;
			*Y = 8112261025743664706L;
		}
	
		/* Synchronize to make sure everyone's buffers are initialized before data transfer is started.  */ 
		PMI_Barrier();	
		
		/* Initialization (T1 == even PEs; T2 == odd PEs). */
		if((pe%2) == 0) {
			status = dmapp_put_nbi(X, seg, target_pe, X, 1, DMAPP_QW);
	                if (status != DMAPP_RC_SUCCESS) {
	                        dmapp_explain_error(status, (const char **)&expl);
	                        fprintf(stderr, " dmapp_put FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
	                        exit(1);
	                }
			status = dmapp_put_nbi(X, seg, target_pe, Y, 1, DMAPP_QW);
	                if (status != DMAPP_RC_SUCCESS) {
	                        dmapp_explain_error(status, (const char **)&expl);
	                        fprintf(stderr, " dmapp_put FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
	                        exit(1);
	                }
			dmapp_gsync_wait();
			status = dmapp_get_nbi(Z, X, seg, target_pe, 1, DMAPP_QW);
	                if (status != DMAPP_RC_SUCCESS) {
	                        dmapp_explain_error(status, (const char **)&expl);
	                        fprintf(stderr, " dmapp_get FAILED: w/ %d - %s -%s \n", status, dmapp_err_str[status], expl);
	                        exit(1);
	                }
			dmapp_gsync_wait();
		} else {
			sleep(1/(random_at_most(100)+1));
			*X = 9123483147965886918L;
	        }
	        
		/* Synchronize before verifying the data. */
		PMI_Barrier();
	
		/* Verify the data. */
		if ((pe%2) == 0) {
			long long c = *Z;
			if (!(c == 9223372036854775807L || c == 8112261025743664706L || c == 9123483147965886918L)) {
				printf("FAILED(Z=%ld);\n", c);
				exit(1);
			}
		}
		
		/* Free buffers allocated from sheap. */
		dmapp_sheap_free(X);
		dmapp_sheap_free(Y);
		dmapp_sheap_free(Z);
	}
	
	/* Release DMAPP resources. This is a mandatory call. */
	status = dmapp_finalize();
	if (status != DMAPP_RC_SUCCESS) {
		dmapp_explain_error(status, (const char **)&expl);
		fprintf(stderr, " dmapp_finalize FAILED w/ %d - %s -%s\n", status, dmapp_err_str[status], expl);
		exit(1);
	}	

	return(0);
}

