#include "rdma-common.h"

struct message {
  enum { 
    MSG_MR,
    MSG_DONE
  } type;

  union {
    struct ibv_mr mr;
  } data;
};

struct context {
  struct ibv_context *ctx;
  struct ibv_pd *pd;
  struct ibv_cq *cq;
  struct ibv_comp_channel *comp_channel;

  pthread_t cq_poller_thread;
};

struct connection {
  struct rdma_cm_id *id;
  struct ibv_qp *qp;

  int connected;

  struct ibv_mr *recv_mr;
  struct ibv_mr *send_mr;
  struct ibv_mr *rdma_mr_X;
  struct ibv_mr *rdma_mr_Y;

  struct ibv_mr peer_mr_X;
  struct ibv_mr peer_mr_Y;

  struct message *recv_msg;
  struct message *send_msg;

  long long *rdma_region_X;
  long long *rdma_region_Y;

  enum {
    SS_INIT,
    SS_MR_SENT_X,
    SS_MR_SENT_Y,
    SS_RDMA_SENT,
    SS_DONE_SENT
  } send_state;

  enum {
    RS_INIT,
    RS_MR_RECV_X,
    RS_MR_RECV_Y,
    RS_DONE_RECV
  } recv_state;

};

static void build_context(struct ibv_context *verbs);
static void build_qp_attr(struct ibv_qp_init_attr *qp_attr);
static void on_completion(struct ibv_wc *);
static void * poll_cq(void *);
static void post_receives(struct connection *conn);
static void register_memory(struct connection *conn);
static void send_message(struct connection *conn);

static void rdma_operation(struct connection *conn, struct ibv_mr peer_mr, long long *rdma_region, struct ibv_mr *rdma_mr, enum ibv_wr_opcode opcode);
static void put_fromYtoY(struct connection *conn);
static void put_fromYtoX(struct connection *conn);
static void put_fromXtoY(struct connection *conn);
static void put_fromXtoX(struct connection *conn);
static void get_fromYtoY(struct connection *conn);
static void get_fromYtoX(struct connection *conn);
static void get_fromXtoY(struct connection *conn);
static void get_fromXtoX(struct connection *conn);

static void test_init(struct connection *conn);
static void test_result(struct connection *conn);
static void test(struct connection *conn);

static struct context *s_ctx = NULL;

static enum app_type app;
static int test_num;

static long long a, b;

void die(const char *reason) {
  fprintf(stderr, "%s\n", reason);
  exit(EXIT_FAILURE);
}

void build_connection(struct rdma_cm_id *id) {
  struct connection *conn;
  struct ibv_qp_init_attr qp_attr;

  build_context(id->verbs);
  build_qp_attr(&qp_attr);

  TEST_NZ(rdma_create_qp(id, s_ctx->pd, &qp_attr));

  id->context = conn = (struct connection *)malloc(sizeof(struct connection));

  conn->id = id;
  conn->qp = id->qp;

  conn->send_state = SS_INIT;
  conn->recv_state = RS_INIT;

  conn->connected = 0;

  register_memory(conn);
  post_receives(conn);
}

void build_context(struct ibv_context *verbs) {
  if (s_ctx) {
    if (s_ctx->ctx != verbs)
      die("cannot handle events in more than one context.");

    return;
  }

  s_ctx = (struct context *)malloc(sizeof(struct context));

  s_ctx->ctx = verbs;

  TEST_Z(s_ctx->pd = ibv_alloc_pd(s_ctx->ctx));
  TEST_Z(s_ctx->comp_channel = ibv_create_comp_channel(s_ctx->ctx));
  TEST_Z(s_ctx->cq = ibv_create_cq(s_ctx->ctx, 10, NULL, s_ctx->comp_channel, 0)); /* cqe=10 is arbitrary */
  TEST_NZ(ibv_req_notify_cq(s_ctx->cq, 0));

  TEST_NZ(pthread_create(&s_ctx->cq_poller_thread, NULL, poll_cq, NULL));
}

void build_params(struct rdma_conn_param *params) {
  memset(params, 0, sizeof(*params));

  params->initiator_depth = params->responder_resources = 1;
  params->rnr_retry_count = 7; /* infinite retry */
}

void build_qp_attr(struct ibv_qp_init_attr *qp_attr) {
  memset(qp_attr, 0, sizeof(*qp_attr));

  qp_attr->send_cq = s_ctx->cq;
  qp_attr->recv_cq = s_ctx->cq;
  qp_attr->qp_type = IBV_QPT_RC;

  qp_attr->cap.max_send_wr = 10;
  qp_attr->cap.max_recv_wr = 10;
  qp_attr->cap.max_send_sge = 1;
  qp_attr->cap.max_recv_sge = 1;
}

void destroy_connection(void *context) {
  struct connection *conn = (struct connection *)context;

  rdma_destroy_qp(conn->id);

  ibv_dereg_mr(conn->send_mr);
  ibv_dereg_mr(conn->recv_mr);
  ibv_dereg_mr(conn->rdma_mr_X);
  ibv_dereg_mr(conn->rdma_mr_Y);

  free(conn->send_msg);
  free(conn->recv_msg);
  free(conn->rdma_region_X);
  free(conn->rdma_region_Y);

  rdma_destroy_id(conn->id);

  free(conn);
}

void on_completion(struct ibv_wc *wc) {
	struct connection *conn = (struct connection *)(uintptr_t)wc->wr_id;
	
	if (wc->status != IBV_WC_SUCCESS) {
		printf("on_completion: status is not IBV_WC_SUCCESS.\n");
        }

	if (wc->opcode & IBV_WC_RECV) {
	        if (conn->recv_state != RS_DONE_RECV) {
			conn->recv_state++;
		}
		
		if (conn->recv_msg->type == MSG_MR) {
			if (conn->recv_state == RS_MR_RECV_X) {
				memcpy(&conn->peer_mr_X, &conn->recv_msg->data.mr, sizeof(conn->peer_mr_X));
			} else if (conn->recv_state == RS_MR_RECV_Y) {
				memcpy(&conn->peer_mr_Y, &conn->recv_msg->data.mr, sizeof(conn->peer_mr_Y));
			} 
			post_receives(conn); /* only rearm for MSG_MR */

			if (conn->send_state == SS_INIT) { /* received peer's MR before sending ours, so send ours back */
				send_mr_X(conn); 
			} else if (conn->send_state == SS_MR_SENT_X) { /* received peer's MR before sending ours, so send ours back */
				send_mr_Y(conn);
			}
		} 
	} else {
		if (conn->send_state != SS_DONE_SENT) {
                	conn->send_state++;
		}
	}
	
	if (conn->send_state == SS_MR_SENT_Y && conn->recv_state == RS_MR_RECV_Y && 											
	   (!((wc->opcode & IBV_WC_RDMA_WRITE) || (wc->opcode & IBV_WC_RDMA_READ) || (wc->opcode & IBV_WR_RDMA_WRITE) || (wc->opcode & IBV_WR_RDMA_READ)))) {
 		test(conn);	
		conn->send_msg->type = MSG_DONE;
		send_message(conn);
	} else if (conn->send_state == SS_DONE_SENT && conn->recv_state == RS_DONE_RECV && 
	          (!((wc->opcode & IBV_WC_RDMA_WRITE) || (wc->opcode & IBV_WC_RDMA_READ) || (wc->opcode & IBV_WR_RDMA_WRITE) || (wc->opcode & IBV_WR_RDMA_READ)))) {
		test_result(conn);
		conn->recv_state = RS_INIT;
        	rdma_disconnect(conn->id);
	}
}		

void on_connect(void *context) {
  ((struct connection *)context)->connected = 1;
}

void * poll_cq(void *ctx) {
  struct ibv_cq *cq;
  struct ibv_wc wc;

  while (1) {
    TEST_NZ(ibv_get_cq_event(s_ctx->comp_channel, &cq, &ctx));
    ibv_ack_cq_events(cq, 1);
    TEST_NZ(ibv_req_notify_cq(cq, 0));

    while (ibv_poll_cq(cq, 1, &wc))
      on_completion(&wc);
  }

  return NULL;
}

void post_receives(struct connection *conn) {
  struct ibv_recv_wr wr, *bad_wr = NULL;
  struct ibv_sge sge;

  wr.wr_id = (uintptr_t)conn;
  wr.next = NULL;
  wr.sg_list = &sge;
  wr.num_sge = 1;

  sge.addr = (uintptr_t)conn->recv_msg;
  sge.length = sizeof(struct message);
  sge.lkey = conn->recv_mr->lkey;

  TEST_NZ(ibv_post_recv(conn->qp, &wr, &bad_wr));
}

void register_memory(struct connection *conn) {
  conn->send_msg = malloc(sizeof(struct message));
  conn->recv_msg = malloc(sizeof(struct message));

  //conn->rdma_region_X = malloc(sizeof(long long));
  //conn->rdma_region_Y = malloc(sizeof(long long));
  posix_memalign((void**)&(conn->rdma_region_X), 64, sizeof(long long));
  posix_memalign((void**)&(conn->rdma_region_Y), 64, sizeof(long long));

  *(conn->rdma_region_X) = 0;
  *(conn->rdma_region_Y) = 0;

  test_init(conn);
  
  TEST_Z(conn->send_mr = ibv_reg_mr(
    s_ctx->pd, 
    conn->send_msg, 
    sizeof(struct message), 
    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ));

  TEST_Z(conn->recv_mr = ibv_reg_mr(
    s_ctx->pd, 
    conn->recv_msg, 
    sizeof(struct message), 
    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ));

  TEST_Z(conn->rdma_mr_X = ibv_reg_mr(
    s_ctx->pd, 
    conn->rdma_region_X, 
    sizeof(long long), 
    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ));

  TEST_Z(conn->rdma_mr_Y = ibv_reg_mr(
    s_ctx->pd, 
    conn->rdma_region_Y, 
    sizeof(long long), 
    IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ));
}

void send_message(struct connection *conn) {
  struct ibv_send_wr wr, *bad_wr = NULL;
  struct ibv_sge sge;

  memset(&wr, 0, sizeof(wr));

  wr.wr_id = (uintptr_t)conn;
  wr.opcode = IBV_WR_SEND;
  wr.sg_list = &sge;
  wr.num_sge = 1;
  wr.send_flags = IBV_SEND_SIGNALED;

  sge.addr = (uintptr_t)conn->send_msg;
  sge.length = sizeof(struct message);
  sge.lkey = conn->send_mr->lkey;

  while (!conn->connected);

  TEST_NZ(ibv_post_send(conn->qp, &wr, &bad_wr));
}

void send_mr_X(void *context) {
  struct connection *conn = (struct connection *)context;

  conn->send_msg->type = MSG_MR;
  memcpy(&conn->send_msg->data.mr, conn->rdma_mr_X, sizeof(struct ibv_mr));

  send_message(conn);
}

void send_mr_Y(void *context) {
  struct connection *conn = (struct connection *)context;

  conn->send_msg->type = MSG_MR;
  memcpy(&conn->send_msg->data.mr, conn->rdma_mr_Y, sizeof(struct ibv_mr));

  send_message(conn);
}

void set_app(enum app_type a) {
  app = a;
}

void set_test_num(int num) {
  test_num = num;
}

void rdma_operation(struct connection *conn, struct ibv_mr peer_mr, long long *rdma_region, struct ibv_mr *rdma_mr, enum ibv_wr_opcode opcode) {
	struct ibv_send_wr wr, *bad_wr = NULL;
	struct ibv_sge sge;
	
	memset(&wr, 0, sizeof(wr));
	
	wr.wr_id = (uintptr_t)conn;
	wr.opcode = opcode;
	wr.sg_list = &sge;
	wr.num_sge = 1;
	wr.send_flags = IBV_SEND_SIGNALED;
	wr.wr.rdma.remote_addr = (uintptr_t)peer_mr.addr;
	wr.wr.rdma.rkey = peer_mr.rkey;
	
	sge.addr = (uintptr_t)rdma_region;
	sge.length = sizeof(long long);
	sge.lkey = rdma_mr->lkey;
	
	TEST_NZ(ibv_post_send(conn->qp, &wr, &bad_wr));
}

void put_fromYtoX(struct connection *conn) {
	rdma_operation(conn, conn->peer_mr_X, conn->rdma_region_Y, conn->rdma_mr_Y, IBV_WR_RDMA_WRITE);
}

void put_fromYtoY(struct connection *conn) {
	rdma_operation(conn, conn->peer_mr_Y, conn->rdma_region_Y, conn->rdma_mr_Y, IBV_WR_RDMA_WRITE);
}

void put_fromXtoY(struct connection *conn) {
	rdma_operation(conn, conn->peer_mr_Y, conn->rdma_region_X, conn->rdma_mr_X, IBV_WR_RDMA_WRITE);
}

void put_fromXtoX(struct connection *conn) {
	rdma_operation(conn, conn->peer_mr_X, conn->rdma_region_X, conn->rdma_mr_X, IBV_WR_RDMA_WRITE);
}

void get_fromYtoX(struct connection *conn) {
	rdma_operation(conn, conn->peer_mr_Y, conn->rdma_region_X, conn->rdma_mr_X, IBV_WR_RDMA_READ);
}

void get_fromYtoY(struct connection *conn) {
        rdma_operation(conn, conn->peer_mr_Y, conn->rdma_region_Y, conn->rdma_mr_Y, IBV_WR_RDMA_READ);
}

void get_fromXtoY(struct connection *conn) {
        rdma_operation(conn, conn->peer_mr_X, conn->rdma_region_Y, conn->rdma_mr_Y, IBV_WR_RDMA_READ);
}

void get_fromXtoX(struct connection *conn) {
        rdma_operation(conn, conn->peer_mr_X, conn->rdma_region_X, conn->rdma_mr_X, IBV_WR_RDMA_READ);
}

/* ****************************** START OF TESTS SECTION ****************************** */

/* ***** START OF TEST_1 ***** */
void test_1_init(struct connection *conn) {
	if (app == CLIENT) {
		*(conn->rdma_region_X) = 0;
		*(conn->rdma_region_Y) = 0;
	} else {
		*(conn->rdma_region_X) = 0;
		*(conn->rdma_region_Y) = 0;
	}
}

void test_1_result(struct connection *conn) {
	if (app == CLIENT) {
		a = *(conn->rdma_region_X);
		b = *(conn->rdma_region_Y);
		if ((a == 1 && b == 0)) { 
			printf("FAILED(%lld,%lld);\n", a, b);
		} else {
			printf("(%lld,%lld)", a, b);
		}
	}
}

void test_1(struct connection *conn) {
        if (app == CLIENT) { 
                get_fromXtoX(conn);
		get_fromXtoY(conn);
        } else {
		*(conn->rdma_region_X) = 1;
	}
}
/* ***** END OF TEST_1 ***** */

/* ***** START OF TEST_2 ***** */
void test_2_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_2_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_X);
                b = *(conn->rdma_region_Y);
                if ((a == 1 && b == 0)) {
                        printf("FAILED(%lld,%lld)\n", a, b);
                } else {
			printf("(%lld,%lld)", a, b);
		}
        }
}

void test_2(struct connection *conn) {
        if (app == CLIENT) {
                get_fromYtoX(conn);
                get_fromXtoY(conn);
        } else {
		*(conn->rdma_region_X) = 1;
		*(conn->rdma_region_Y) = 1;
        }
}
/* ***** END OF TEST_2 ***** */

/* ***** START OF TEST_3 ***** */
void test_3_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_3_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_X);
                if (!(a == 0 || a == 3)) {
                        printf("FAILED(%lld)\n", a);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_3(struct connection *conn) {
        if (app == CLIENT) {
                get_fromXtoX(conn);
                get_fromYtoX(conn);
        } else {
                *(conn->rdma_region_Y) = 3;
                *(conn->rdma_region_X) = 2;
        }
}
/* ***** END OF TEST_3 ***** */

/* ***** START OF TEST_4 (FAILED) ***** */
void test_4_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_4_result(struct connection *conn) {
        if (app == CLIENT) {
	        a = *(conn->rdma_region_X);
                if (!(a == 1)) {
                        printf("FAILED(%lld, old(a)=%lld)\n", a, b);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_4(struct connection *conn) {
        if (app == CLIENT) {
		*(conn->rdma_region_X) = 0;
		*(conn->rdma_region_Y) = 1;
		put_fromXtoX(conn);
		put_fromYtoX(conn);
		get_fromXtoX(conn);
		b = *(conn->rdma_region_X);
        } 
}
/* ***** END OF TEST_4 ***** */

/* ***** START OF TEST_5 (FAILED) ***** */
void test_5_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_5_result(struct connection *conn) {
        if (app == CLIENT) {
	        a = *(conn->rdma_region_X);
                if (!(a == 0)) {
			printf("FAILED(%lld)", a);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_5(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 1;
                get_fromXtoX(conn);
                put_fromXtoX(conn);	
		get_fromXtoX(conn);
        }
}
/* ***** END OF TEST_5 ***** */

/* ***** START OF TEST_6 (FAILED) ***** */
void test_6_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_6_result(struct connection *conn) {
        if (app == CLIENT) {
        	a = *(conn->rdma_region_X);
                if (!(a == 0)) {
                        printf("FAILED(%lld)", a);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_6(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 1;
                get_fromYtoX(conn);
                put_fromXtoX(conn);
		get_fromXtoX(conn);
        } 
}
/* ***** END OF TEST_6 ***** */

/* ***** START OF TEST_7 ***** */
void test_7_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_7_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_Y);
                if (!(a == 1)) {
                        printf("FAILED(%lld)\n", a);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_7(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 1;
                put_fromXtoX(conn);
                get_fromXtoY(conn);
        } else {
        }
}
/* ***** END OF TEST_7 ***** */

/* ***** START OF TEST_8 ***** */
void test_8_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_8_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_X);
                if (!(a == 1)) {
                        printf("FAILED(%lld)\n", a);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_8(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 1;
                put_fromXtoX(conn);
                get_fromXtoX(conn);
        } else {
        }
}
/* ***** END OF TEST_8 ***** */

/* ***** START OF TEST_9 ***** */
void test_9_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_9_result(struct connection *conn) {
	if (app == CLIENT) {
                a = *(conn->rdma_region_X);
                if (!(a == 1)) {
                        printf("FINAL_FAILED(%lld)\n", a);
                } 
        }
}

void test_9(struct connection *conn) {
        if (app == SERVER) {
                *(conn->rdma_region_Y) = 1;
                put_fromYtoX(conn);
        } else {
		a = *(conn->rdma_region_X);
		while (a == 0) {
			a = *(conn->rdma_region_X);
		}
		if (!(a == 1)) {
                        printf("PRE_FAILED(%lld)\n", a);
              	}
        }
}
/* ***** END OF TEST_9 ***** */

/* ***** START OF TEST_10 ***** */
void test_10_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_10_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1)) {
                        printf("FAILED(%lld)\n", a);
                } else {
			printf("(%lld)", a);
		}
        }
}

void test_10(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_Y) = 1;
		get_fromXtoY(conn);
		a = *(conn->rdma_region_Y);
        } else {
		*(conn->rdma_region_X) = 0;
        }
}
/* ***** END OF TEST_10 ***** */

/* ***** START OF TEST_11 ***** */
void test_11_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_11_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_Y);
		if (!(a == 0 || a == 1)) {
                        printf("FAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_11(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_Y) = 1;
                get_fromXtoY(conn);
        } else {
                *(conn->rdma_region_X) = 0;
		*(conn->rdma_region_X) = 1;
        }
}
/* ***** END OF TEST_11 ***** */

/* ***** START OF TEST_12 ***** */
void test_12_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_12_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_Y);
                if (!(a == 0 || a == 1)) {
                        printf("FAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_12(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_Y) = 1;
                put_fromYtoX(conn);
		*(conn->rdma_region_Y) = 1;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_X) = 1;
        }
}
/* ***** END OF TEST_12 ***** */

/* ***** START OF TEST_13 ***** */
void test_13_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_13_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_13(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                a = *(conn->rdma_region_X);
        } else {
                *(conn->rdma_region_Y) = 1;
                put_fromYtoX(conn);
        }
}
/* ***** END OF TEST_13 ***** */

/* ***** START OF TEST_14 ***** */
void test_14_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_14_result(struct connection *conn) {
        if (app == CLIENT) {
		a = *(conn->rdma_region_Y);
                if (!(a == 0 || a == 1 || a == 2)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_14(struct connection *conn) {
        if (app == CLIENT) {
		*(conn->rdma_region_Y) = 1;
		put_fromYtoX(conn);
		get_fromXtoY(conn);
        } else {
                *(conn->rdma_region_X) = 0;
		*(conn->rdma_region_X) = 2;
        }
}
/* ***** END OF TEST_14 ***** */

/* ***** START OF TEST_15 ***** */
void test_15_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_15_result(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_Y);
                if (!(a == 0 || a == 1 || a == 3)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_15(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_Y) = 0;
                *(conn->rdma_region_X) = 1;
		put_fromXtoX(conn);
		get_fromXtoY(conn);
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_X) = 3;
        }
}
/* ***** END OF TEST_15 ***** */

/* ***** START OF TEST_16 ***** */
void test_16_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_16_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1 || a == 5)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_16(struct connection *conn) {
        if (app == SERVER) {
                *(conn->rdma_region_Y) = 0;
                *(conn->rdma_region_X) = 1;
                put_fromXtoX(conn);
                get_fromXtoY(conn);
		*(conn->rdma_region_X) = 5;
		put_fromXtoX(conn);
        } else {
                *(conn->rdma_region_X) = 0;
                a = *(conn->rdma_region_X);
		test_16_result(conn);
        }
}
/* ***** END OF TEST_16 ***** */

/* ***** START OF TEST_17 ***** */
void test_17_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_17_result(struct connection *conn) {
        if (app == CLIENT) {
	        if (!(a == 0 || a == 1)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_17(struct connection *conn) {
        if (app == CLIENT) {
		*(conn->rdma_region_X) = 0;
                *(conn->rdma_region_X) = 1;
                a = *(conn->rdma_region_X);
		test_17_result(conn);
        } else {
                *(conn->rdma_region_Y) = 0;
                *(conn->rdma_region_Y) = 2;
                get_fromXtoY(conn);
                put_fromYtoX(conn);
        }
}
/* ***** END OF TEST_17 ***** */

/* ***** START OF TEST_18 ***** */
void test_18_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_18_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1 || a == 2)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_18(struct connection *conn) {
        if (app == SERVER) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_X) = 1;
		put_fromXtoY(conn);
		get_fromYtoX(conn);
        } else {
                *(conn->rdma_region_Y) = 0;
		*(conn->rdma_region_Y) = 2;
		get_fromXtoY(conn);
		put_fromYtoX(conn);
		a = *(conn->rdma_region_Y);
        }
}
/* ***** END OF TEST_18 ***** */

/* ***** START OF TEST_19 ***** */
void test_19_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_19_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1 || a == 2)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_19(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
		put_fromXtoY(conn);
                *(conn->rdma_region_X) = 1;
                put_fromXtoY(conn);
                get_fromYtoX(conn);
		a = *(conn->rdma_region_X);
        } else {
		*(conn->rdma_region_Y) = 2;
        }
}
/* ***** END OF TEST_19 ***** */

/* ***** START OF TEST_20 ***** */
void test_20_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_20_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_20(struct connection *conn) {
        if (app == CLIENT) {
		get_fromXtoX(conn);
                *(conn->rdma_region_X) = 1;
                a = *(conn->rdma_region_X);
        } else {
                *(conn->rdma_region_X) = 0;
        }
}
/* ***** END OF TEST_20 ***** */

/* ***** START OF TEST_21 ***** */
void test_21_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_21_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!(a == 0 || a == 1)) {
                        printf("\nFAILED(%lld);\n", a);
                } else {
                        printf("(%lld)", a);
                }
        }
}

void test_21(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 1;
                a = *(conn->rdma_region_X);
		test_21_result(conn);
        } else {
                *(conn->rdma_region_X) = 0;
		put_fromXtoX(conn);
        }
}
/* ***** END OF TEST_21 ***** */

/* ***** START OF TEST_22 ***** */
void test_22_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_22_result(struct connection *conn) {
        if (app == CLIENT) {
		a = *(conn->rdma_region_Y);
		b = *(conn->rdma_region_X);
                if (!((a == 0 && b == 0) || (a == 0 && b == 1) || (a == 1 && b == 1))) {
                        printf("\nFAILED(%lld,%lld);\n", a, b);
                }
        }
}

void test_22(struct connection *conn) {
        if (app == CLIENT) {
                get_fromYtoY(conn);
        } else {
                *(conn->rdma_region_X) = 1;
		put_fromXtoX(conn);
		*(conn->rdma_region_Y) = 1;
        }
}
/* ***** END OF TEST_22 ***** */

/* ***** START OF TEST_23 ***** */
void test_23_init(struct connection *conn) {
        if (app == CLIENT) {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        } else {
                *(conn->rdma_region_X) = 0;
                *(conn->rdma_region_Y) = 0;
        }
}

void test_23_result(struct connection *conn) {
        if (app == CLIENT) {
                if (!((a == 0 && b == 0))) {
                        printf("\nFAILED(%lld,%lld);\n", a, b);
                }
        }
}

void test_23(struct connection *conn) {
        if (app == CLIENT) {
                a = *(conn->rdma_region_Y);
		get_fromYtoY(conn);
                b = *(conn->rdma_region_X);
		get_fromXtoX(conn);
        } else {
                *(conn->rdma_region_X) = 1;
                *(conn->rdma_region_Y) = 1;
		put_fromYtoY(conn); //dummy call due to the static declaration
        }
}
/* ***** END OF TEST_23 ***** */

/* ***** START OF TEST_24 ***** */
void test_24_init(struct connection *conn) {
	if (app == CLIENT) {
                *(conn->rdma_region_X) = 9223372036854775807L;
                *(conn->rdma_region_Y) = 8112261025733664706L;
        } else {
                *(conn->rdma_region_X) = 0L;
                *(conn->rdma_region_Y) = 0L;
        }
}

void test_24_result(struct connection *conn) {
        if (app == CLIENT) {
        	if (!(*(conn->rdma_region_X) == 9223372036854775807L || *(conn->rdma_region_X) == 8112261025733664706L)) {
                	printf("FAILED(%lld);", *(conn->rdma_region_X));
         	}
        }
}

void test_24(struct connection *conn) {
        if (app == CLIENT) {
        	put_fromXtoX(conn);
		put_fromYtoX(conn);
		get_fromXtoX(conn);		
        }
}
/* ***** END OF TEST_24 ***** */

void test_init(struct connection *conn) {
	switch(test_num) {
	case 1 :
		test_1_init(conn);
		break;  
	case 2:
		test_2_init(conn);
		break;
	case 3:
		test_3_init(conn);
		break;
	case 4:
		test_4_init(conn);
		break;
	case 5:
		test_5_init(conn);
		break;
	case 6:	
		test_6_init(conn);
		break;
	case 7:
		test_7_init(conn);
		break;
	case 8:
		test_8_init(conn);
		break;
	case 9:
		test_9_init(conn);
		break;
	case 10:
		test_10_init(conn);
		break;
	case 11:
		test_11_init(conn);
		break;
	case 12:
		test_12_init(conn);
		break;
	case 13:
		test_13_init(conn);
		break;
	case 14:
		test_14_init(conn);
		break;
	case 15:
		test_15_init(conn);
		break;
	case 16:
		test_16_init(conn);
		break;
	case 17:
		test_17_init(conn);
		break;
	case 18:
		test_18_init(conn);
		break;
	case 19:
		test_19_init(conn);
		break;
	case 20:
		test_20_init(conn);
		break;
	case 21:
		test_21_init(conn);
		break;
	case 22:
		test_22_init(conn);
		break;
	case 23:
		test_23_init(conn);
		break;
	case 24:
		test_24_init(conn);
		break;
	default : 
		break;
	}
}

void test_result(struct connection *conn) {
	switch(test_num) {
	case 1 :
		test_1_result(conn);
		break;
	case 2:
		test_2_result(conn);
		break;
	case 3:
		test_3_result(conn);
		break;
	case 4:
		test_4_result(conn);
		break;
	case 5:	
		test_5_result(conn);
		break;
	case 6:
		test_6_result(conn);
		break;
	case 7:
		test_7_result(conn);
		break;
	case 8:
		test_8_result(conn);
		break;
	case 9:
		test_9_result(conn);
		break;
	case 10:
		test_10_result(conn);
		break; 
	case 11:	
		test_11_result(conn);
		break;
	case 12:
		test_12_result(conn);
		break;
	case 13:
		test_13_result(conn);
		break;
	case 14:
		test_14_result(conn);
		break;
	case 15:
		test_15_result(conn);
		break;
	case 16:
		test_16_result(conn);
		break;
	case 17:
		test_17_result(conn);
		break;
	case 18:
		test_18_result(conn);
		break;
	case 19:
		test_19_result(conn);
		break;
	case 20:
		test_20_result(conn);
		break;
	case 21:
		test_21_result(conn);
		break;
	case 22:
		test_22_result(conn);
		break;
	case 23:
		test_23_result(conn);
		break;
	case 24:
		test_24_result(conn);
		break;
	default : 
		break;
	}
}

void test(struct connection *conn) {
	switch(test_num){
	case 1 :
		test_1(conn);     
		break;
	case 2 :
		test_2(conn);
		break;
	case 3:
		test_3(conn);
		break;
	case 4:
		test_4(conn);
		break;
	case 5:
		test_5(conn);
		break;
	case 6:
		test_6(conn);
		break;
	case 7:
		test_7(conn);
		break;
	case 8:
		test_8(conn);
		break;
	case 9:
		test_9(conn);
		break;
	case 10:
		test_10(conn);
		break;
	case 11:
		test_11(conn);
		break;
	case 12:
		test_12(conn);
		break;
	case 13:
		test_13(conn);
		break;
	case 14:
		test_14(conn);
		break;
	case 15:
		test_15(conn);
		break;
	case 16:
		test_16(conn);
		break;
	case 17:
		test_17(conn);
		break;
	case 18:
		test_18(conn);
		break;
	case 19:
		test_19(conn);
		break;
	case 20:
		test_20(conn);
		break;
	case 21:
		test_21(conn);
		break;
	case 22:
		test_22(conn);
		break;
	case 23:
		test_23(conn);
		break;
	case 24:
		test_24(conn);
		break;
	default :
		break;
	}
}

/* ****************************** END OF TESTS SECTION ****************************** */

